@@ -154,10 +154,8 @@ device_selected_cb(GObject *object,
 	char *address;
 	gboolean connected;
 
-	g_object_get (G_OBJECT (adapter->chooser),
-		      "device-selected", &address,
-		      "device-selected-is-connected", &connected,
-		      NULL);
+	address = bluetooth_chooser_get_selected_device (BLUETOOTH_CHOOSER (adapter->chooser));
+	connected = bluetooth_chooser_get_selected_device_is_connected (BLUETOOTH_CHOOSER (adapter->chooser));
 	update_buttons(adapter, (address != NULL), connected);
 }
 
@@ -202,10 +200,16 @@ static gboolean show_confirm_dialog(void)
 static void delete_callback(GtkWidget *button, gpointer user_data)
 {
 	adapter_data *adapter = user_data;
+	GValue value = { 0, };
 	DBusGProxy *device;
 	const char *path;
 
-	g_object_get (G_OBJECT (adapter->chooser), "device-selected-proxy", &device, NULL);
+	if (bluetooth_chooser_get_selected_device_info (BLUETOOTH_CHOOSER (adapter->chooser),
+							"proxy", &value) == FALSE) {
+		return;
+	}
+	device = g_value_dup_object (&value);
+	g_value_unset (&value);
 
 	if (device == NULL)
 		return;
@@ -224,9 +228,15 @@ static void delete_callback(GtkWidget *button, gpointer user_data)
 static void disconnect_callback(GtkWidget *button, gpointer user_data)
 {
 	adapter_data *adapter = user_data;
+	GValue value = { 0, };
 	DBusGProxy *device;
 
-	g_object_get (G_OBJECT (adapter->chooser), "device-selected-proxy", &device, NULL);
+	if (bluetooth_chooser_get_selected_device_info (BLUETOOTH_CHOOSER (adapter->chooser),
+							"proxy", &value) == FALSE) {
+		return;
+	}
+	device = g_value_dup_object (&value);
+	g_value_unset (&value);
 
 	if (device == NULL)
 		return;
@@ -257,6 +267,7 @@ static void create_adapter(adapter_data *adapter)
 {
 	GHashTable *hash = NULL;
 	GValue *value;
+	DBusGProxy *default_proxy;
 	const gchar *address, *name;
 	gboolean powered, discoverable;
 	guint timeout;
@@ -304,6 +315,13 @@ static void create_adapter(adapter_data *adapter)
 	adapter->discoverable = discoverable;
 	adapter->timeout_value = timeout;
 
+	default_proxy = bluetooth_client_get_default_adapter (client);
+	if (default_proxy != NULL) {
+		adapter->is_default = g_str_equal (dbus_g_proxy_get_path (default_proxy),
+						   dbus_g_proxy_get_path (adapter->proxy));
+		g_object_unref (default_proxy);
+	}
+
 	mainbox = gtk_vbox_new(FALSE, 6);
 	gtk_container_set_border_width(GTK_CONTAINER(mainbox), 12);
 
@@ -387,8 +405,6 @@ static void create_adapter(adapter_data *adapter)
 
 	g_signal_connect (adapter->chooser, "notify::device-selected",
 			  G_CALLBACK(device_selected_cb), adapter);
-	g_signal_connect (adapter->chooser, "notify::device-selected-is-connected",
-			  G_CALLBACK(device_selected_cb), adapter);
 
 	gtk_table_attach(GTK_TABLE(table), adapter->chooser, 0, 1, 1, 2,
 			 GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
@@ -510,25 +526,18 @@ static adapter_data *adapter_alloc(GtkTreeModel *model,
 {
 	DBusGProxy *proxy;
 	adapter_data *adapter;
-	gboolean is_default;
-
-	adapter = g_new0(adapter_data, 1);
 
 	gtk_tree_model_get(model, iter,
 			   BLUETOOTH_COLUMN_PROXY, &proxy,
-			   BLUETOOTH_COLUMN_DEFAULT, &is_default,
 			   -1);
 
-	if (proxy == NULL) {
-		g_free(adapter);
+	if (proxy == NULL)
 		return NULL;
-	}
 
+	adapter = g_new0(adapter_data, 1);
 	adapter->notebook = user_data;
-
 	adapter->reference = gtk_tree_row_reference_new(model, path);
 	adapter->proxy = proxy;
-	adapter->is_default = is_default;
 
 	return adapter;
 }
@@ -631,9 +640,6 @@ adapter_changed (GtkTreeModel *model,
 			   BLUETOOTH_COLUMN_NAME, &name,
 			   -1);
 
-	if (proxy == NULL)
-		return;
-
 	for (i = 0; i < count; i++) {
 		GtkWidget *widget;
 		adapter_data *adapter;
@@ -646,14 +652,18 @@ adapter_changed (GtkTreeModel *model,
 		if (adapter == NULL)
 			continue;
 
-		if (proxy == adapter->proxy) {
+		adapter->is_default = is_default;
+
+		if (proxy == NULL || adapter->proxy == NULL)
+			continue;
+
+		if (g_str_equal (dbus_g_proxy_get_path (proxy), dbus_g_proxy_get_path (adapter->proxy)) != FALSE) {
 			if (is_default != FALSE && powered != FALSE)
 				gtk_notebook_set_current_page (notebook, i);
 			/* We usually get an adapter_added before the device
 			 * is powered, so we set the name here instead */
 			if (name)
 				gtk_entry_set_text(GTK_ENTRY(adapter->entry), name);
-			break;
 		}
 	}
 
